;;;; -*- lisp -*-

;;;; -*- lisp -*-

(in-package :it.bese.tar)

(defun read-tar (filename)
  (with-open-file (tar filename
		       :direction :input
		       :element-type '(unsigned-byte 8))
    (assert (not (zerop (file-length tar)))
	    (tar)
	    "Um, so ~S is a tar, but it's empty...yeah, right." filename)
    (let ((files (loop while (read-tar-file tar) collect it)))
      ;; ok, read all the files, now make sure there's one more empty
      ;; zero block (header-tar-header read the second ot last one
      (let ((buf (make-array 512 :element-type '(unsigned-byte 8)))
	    (pos (read-sequence buf tar)))
	(if (and (= pos 512)
		 (every #'zerop buf))
	    (return-from read-tar files)
	    (progn
	      (warning "Tar file ended before last empty block")
	      (return-from read-tar files)))))))

(defun read-tar-header (stream)
  "Read a TAR header from STREAM."
  (let* ((header (make-array 512 :element-type '(unsigned-byte 8)))
	 (pos (read-sequence header stream))
	 (file (make-instance 'tar-file)))
    (assert (= 512 pos)
	    (pos)
	    "File ended before reading complete header.")
    (when (every #'zerop header)
      ;; last block, no header here
      (return-from read-tar-header nil))
    (flet ((string-from-header-subseq (start length)
             (string-from-array (make-array length
					    :element-type '(unsigned-byte 8)
					    :displaced-to header
					    :displaced-index-offset start))))
      (with-slots (name mode uid gid size mtime chksum typeflag linkname magic version
		   uname gname devmajor devminor prefix)
	  file
	(setf name     (string-from-header-subseq 0 100)
	      mode     (string-from-header-subseq 100 8)
	      uid      (string-from-header-subseq 108 8)
	      gid      (string-from-header-subseq 116 8)
	      size     (string-from-header-subseq 124 12)
	      mtime    (string-from-header-subseq 136 12)
	      chksum   (string-from-header-subseq 148 8)
	      typeflag (string-from-header-subseq 156 1)
	      linkname (string-from-header-subseq 157 100)
	      magic    (string-from-header-subseq 257 6)
	      version  (string-From-Header-Subseq 263 2)
	      uname    (string-from-header-subseq 265 32)
	      gname    (string-from-header-subseq 297 32)
	      devmajor (string-from-header-subseq 329 8)
	      devminor (string-from-header-subseq 337 8)
	      prefix   (string-from-header-subseq 345 155))
	;; now convert the numbers into ... numbers
	(macrolet ((parse-octalf (octal-string)
		     `(setf ,octal-string (parse-integer ,octal-string :radix 8))))
	  (parse-octalf mode)
	  (parse-octalf uid)
	  (parse-octalf gid)
	  (parse-octalf size)
	  (parse-octalf mtime)
	  (when (not (string= "" chksum))
	    (parse-octalf chksum))
	  (if (string= "ustar" (subseq magic 0 5))
	      (progn
		(setf name (concatenate 'string prefix name))
		(setf typeflag (eswitch (typeflag :test #'string=)
  			         (("0" "") :regular)
				 ("1" :link)
				 ("2" :symbolic)
				 ("3" :char-special)
				 ("4" :block-special)
				 ("5" :directory)
				 ("6" :fifo-special)))
		(if (and (< 0 (length version))
			 (every #'digit-char-p version))
		    (parse-octalf version)
		    (setf version nil))
		(if (and (< 0 (length devmajor))
			 (every #'digit-char-p devmajor))
		    (parse-octalf devmajor)
		    (setf devmajor nil))
		(if (and (< 0 (length devminor))
			 (every #'digit-char-p devminor))
		    (parse-octalf devminor)
		    (setf devminor nil)))
	      (setf typeflag (eswitch (typeflag :test #'string=)
		               ("l" :link)
			       ("s" :symbolic)
			       ("o" :regular)))))))
    file))

(defun read-tar-file (stream)
  (let ((file (read-tar-header stream)))
    (when (null file)
      (return-from read-tar-file nil))
    (let* ((size (tar-file.size file))
	   (buffer-size (* 512 (ceiling size 512)))
	   (buffer (make-array buffer-size :element-type '(unsigned-byte 8)))
	   (pos (read-sequence buffer stream)))
      (assert (= pos buffer-size)
	      (stream)
	      "Stream ended before reading entire file")
      (setf (tar-file.data file) (make-array size
					     :element-type '(unsigned-byte 8)
					     :displaced-to buffer))
      file)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Copyright (c) 2002-2003, Edward Marco Baringer
;;;; All rights reserved. 
;;;; 
;;;; Redistribution and use in source and binary forms, with or without
;;;; modification, are permitted provided that the following conditions are
;;;; met:
;;;; 
;;;;  - Redistributions of source code must retain the above copyright
;;;;    notice, this list of conditions and the following disclaimer.
;;;; 
;;;;  - Redistributions in binary form must reproduce the above copyright
;;;;    notice, this list of conditions and the following disclaimer in the
;;;;    documentation and/or other materials provided with the distribution.
;;;;
;;;;  - Neither the name of Edward Marco Baringer, nor BESE, nor the names
;;;;    of its contributors may be used to endorse or promote products
;;;;    derived from this software without specific prior written permission.
;;;; 
;;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;; A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
;;;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;;;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;;;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;;;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;;;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;;;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
